"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest
import os
import configsuite


class TestLicenseHeaders(unittest.TestCase):

    header = """\"\"\"Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
\"\"\"
"""

    def setUp(self):
        ignore_extensions = ('.pyc',)

        source_dirs = (
            os.path.dirname(configsuite.__file__),
            os.path.dirname(__file__),
        )

        source_files = [
                os.path.join(os.path.dirname(__file__), '..', 'setup.py')]
        for sdir in source_dirs:
            for path, _, filenames in os.walk(sdir):
                source_files += [ os.path.join(path, fn) for fn in filenames ]

        def is_source_file(filename):
            return (
                os.path.splitext(filename)[1] not in ignore_extensions
                and os.path.basename(filename) != '_version.py'
            )

        self.source_files = filter(
                is_source_file,
                map(os.path.realpath, source_files)
                )


    def test_license_headers(self):
        header_length = self.header.count('\n')

        for sfile in self.source_files:
            with open(sfile) as f:
                file_header = ''.join(f.readlines()[:header_length])
                err_msg = '{} missing correct header'.format(sfile)
                self.assertEqual(self.header, file_header, err_msg)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import numbers
import unittest


import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


def _build_name_pet_schema():
    return {
        MK.Type: types.NamedDict,
        MK.Content: {
            'name': {
                MK.Type: types.String,
            },
            'pet': {
                MK.Type: types.NamedDict,
                MK.Content: {
                    'name': {
                        MK.Type: types.String,
                    },
                    'favourite_food': {
                        MK.Type: types.String,
                    },
                    'weight': {
                        MK.Type: types.Number,
                    },
                    'nkids': {
                        MK.Type: types.Integer,
                    },
                    'likeable': {
                        MK.Type: types.Bool,
                    },
                },
            },
            'playgrounds': {
                MK.Type: types.List,
                MK.Content: {
                    MK.Item: {
                        MK.Type: types.NamedDict,
                        MK.Content: {
                            'name': {
                                MK.Type: types.String,
                            },
                            'score': {
                                MK.Type: types.Integer,
                            },
                        },
                    },
                },
            },
        },
    }


def _build_valid_pet_config():
    return {
        'name': 'Markus',
        'pet': {
            'name': 'Donkey Kong',
            'favourite_food': 'bananas',
            'weight': 1.9,
            'nkids': 8,
            'likeable': True,
        },
        'playgrounds': [],
    }


class TestTypes(unittest.TestCase):

    def test_name_pet_accepted(self):
        raw_config = _build_valid_pet_config()
        config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

        self.assertTrue(config_suite.valid)

        config = config_suite.snapshot
        self.assertEqual(raw_config['name'], config.name)
        self.assertEqual(raw_config['pet']['name'], config.pet.name)
        self.assertEqual(
                raw_config['pet']['favourite_food'],
                config.pet.favourite_food,
                )

    def test_invalid_string(self):
        for name in [14, None, [], (), {}]:
            raw_config = _build_valid_pet_config()
            raw_config['name'] = name
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

            self.assertFalse(config_suite.valid)

            self.assertEqual(1, len(config_suite.errors))
            err = config_suite.errors[0]
            self.assertIsInstance(err, configsuite.InvalidTypeError)
            self.assertEqual(('name',), err.key_path)

            config = config_suite.snapshot
            self.assertEqual(raw_config['name'], config.name)
            self.assertEqual(raw_config['pet']['name'], config.pet.name)
            self.assertEqual(
                    raw_config['pet']['favourite_food'],
                    config.pet.favourite_food,
                    )

    def test_invalid_base_dict(self):
        for raw_config in [14, None, [{'name': 'Markus'}], ()]:
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())
            self.assertFalse(config_suite.valid)

            self.assertEqual(1, len(config_suite.errors))
            err = config_suite.errors[0]
            self.assertIsInstance(err, configsuite.InvalidTypeError)
            self.assertEqual((), err.key_path)

    def test_invalid_dict(self):
        for pet in [14, None, [{'name': 'Markus'}], ()]:
            raw_config = _build_valid_pet_config()
            raw_config['pet'] = pet
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())
            self.assertFalse(config_suite.valid)

            self.assertEqual(1, len(config_suite.errors))
            err = config_suite.errors[0]
            self.assertIsInstance(err, configsuite.InvalidTypeError)
            self.assertEqual(('pet',), err.key_path)

    def test_string(self):
        for strval in ['fdnsjk', '', 'str4thewin', True, [], 1, 1.2, {}, None]:
            raw_config = _build_valid_pet_config()
            raw_config['name'] = strval
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

            if isinstance(strval, str):
                self.assertTrue(config_suite.valid)
                self.assertEqual(strval, config_suite.snapshot.name)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(('name',), err.key_path)

    def test_int(self):
        for intval in ['fdnsjk', '', False, 0, -10, [], 1, 1.2, {}, None]:
            raw_config = _build_valid_pet_config()
            raw_config['pet']['nkids'] = intval
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

            if isinstance(intval, int):
                self.assertTrue(config_suite.valid)
                self.assertEqual(intval, config_suite.snapshot.pet.nkids)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(('pet', 'nkids'), err.key_path)

    def test_number(self):
        for numval in [False, True, 'fdnsjk', '', 0, -10, [], 1, 1.2, {}, None]:
            raw_config = _build_valid_pet_config()
            raw_config['pet']['weight'] = numval
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

            if isinstance(numval, numbers.Number):
                self.assertTrue(config_suite.valid)
                self.assertEqual(numval, config_suite.snapshot.pet.weight)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(('pet', 'weight'), err.key_path)

    def test_bool(self):
        for boolval in [False, True, 'fdnsjk', '', 0, -10, [], 1, 1.2, {}, None]:
            raw_config = _build_valid_pet_config()
            raw_config['pet']['likeable'] = boolval
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

            if isinstance(boolval, bool):
                self.assertTrue(config_suite.valid)
                self.assertEqual(boolval, config_suite.snapshot.pet.likeable)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(('pet', 'likeable'), err.key_path)

    def test_list(self):
        playgrounds = [
            { 'name': 'superfun', 'score': 1 },
            { 'name': 'Hell', 'score': 99 },
        ]

        for end_idx, _ in enumerate(playgrounds):
            raw_config = _build_valid_pet_config()
            raw_config['playgrounds'] = playgrounds[:end_idx]
            config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

            self.assertTrue(config_suite.valid, 'Errors: %r' % (config_suite.errors,))
            config = config_suite.snapshot
            self.assertEqual(end_idx, len(config.playgrounds))

            for idx, plgr in enumerate(playgrounds[:end_idx]):
                self.assertEqual(plgr['name'], config.playgrounds[idx].name)
                self.assertEqual(plgr['score'], config.playgrounds[idx].score)

    def test_list_errors(self):
        raw_config = _build_valid_pet_config()
        raw_config['playgrounds'] = [{ 'name': 'faulty grounds' }]
        config_suite = configsuite.ConfigSuite(raw_config, _build_name_pet_schema())

        self.assertFalse(config_suite.valid)
        self.assertEqual(1, len(config_suite.errors))
        err = config_suite.errors[0]
        self.assertIsInstance(err, configsuite.MissingKeyError)
        self.assertEqual(('playgrounds', 0), err.key_path)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import copy
import configsuite
import collections


from .schema import assert_valid_schema
from .meta_keys import MetaKeys as MK


class ConfigSuite(object):

    def __init__(self, raw_config, schema):
        assert_valid_schema(schema)
        self._raw_config = copy.deepcopy(raw_config)
        self._schema = copy.deepcopy(schema)
        self._valid = None
        self._errors = None
        self._snapshot = None

    @property
    def valid(self):
        self._ensure_validation()
        return self._valid

    @property
    def errors(self):
        self._ensure_validation()
        return self._errors

    @property
    def snapshot(self):
        if self._snapshot is None:
            self._snapshot = self._build_snapshot(
                    self._raw_config,
                    self._schema,
                    )

        return self._snapshot

    def _build_snapshot(self, config, schema):
        if config is None:
            return None

        data_type = schema[MK.Type]
        if isinstance(data_type, configsuite.BasicType):
            return config
        elif data_type == configsuite.types.NamedDict:
            content_schema = schema[MK.Content]
            dict_name = data_type.name
            dict_keys = content_schema.keys()
            dict_collection = collections.namedtuple(dict_name, dict_keys)

            return dict_collection(**{
                key: self._build_snapshot(config.get(key), content_schema[key])
                for key in content_schema
            })
        elif data_type == configsuite.types.List:
            item_schema = schema[MK.Content][MK.Item]
            return tuple(map(
                lambda elem: self._build_snapshot(elem, item_schema),
                config
            ))
        elif data_type == configsuite.types.Dict:
            key_schema = schema[MK.Content][MK.Key]
            value_schema = schema[MK.Content][MK.Value]

            Pair = collections.namedtuple('KeyValuePair', ['key', 'value'])
            return tuple([
                Pair(
                    self._build_snapshot(key, key_schema),
                    self._build_snapshot(value, value_schema)
                )
                for key, value in config.items()
            ])
        else:
            msg = 'Encountered unknown type {} while building snapshot'
            raise TypeError(msg.format(str(data_type)))

    def _ensure_validation(self):
        if self._valid is None:
            validator = configsuite.Validator(self._schema)
            val_res = validator.validate(self._raw_config)
            self._valid = val_res.valid
            self._errors = val_res.errors
